@@ -1,3 +1,5 @@
+// Linux
+
 #[cfg(target_os = "linux")]
 /// Check if `CAP_NET_RAW` is in the permitted set and if so raise it to the effective set.
 pub fn ensure_caps() -> anyhow::Result<()> {
@@ -17,7 +19,9 @@ pub fn drop_caps() -> anyhow::Result<()> {
     Ok(())
 }
 
-#[cfg(not(target_os = "linux"))]
+// macOS, BSD etc
+
+#[cfg(all(unix, not(target_os = "linux")))]
 #[allow(clippy::unnecessary_wraps)]
 /// Ensure the effective user is `root`.
 pub fn ensure_caps() -> anyhow::Result<()> {
@@ -28,7 +32,7 @@ pub fn ensure_caps() -> anyhow::Result<()> {
     Ok(())
 }
 
-#[cfg(not(target_os = "linux"))]
+#[cfg(all(unix, not(target_os = "linux")))]
 #[allow(clippy::unnecessary_wraps)]
 /// Drop all capabilities.
 ///
@@ -36,3 +40,20 @@ pub fn ensure_caps() -> anyhow::Result<()> {
 pub fn drop_caps() -> anyhow::Result<()> {
     Ok(())
 }
+
+// Windows
+
+#[cfg(not(unix))]
+#[allow(clippy::unnecessary_wraps)]
+/// Ensure the effective user is `root`.
+pub fn ensure_caps() -> anyhow::Result<()> {
+    Ok(())
+}
+
+#[cfg(not(unix))]
+/// Drop all capabilities.
+///
+/// This is a no-op on non-Linux systems.
+pub fn drop_caps() -> anyhow::Result<()> {
+    Ok(())
+}
