@@ -87,8 +87,9 @@ pub enum DnsResolveMethod {
 #[derive(Parser, Debug)]
 #[clap(author, version, about, long_about = None)]
 pub struct Args {
-    /// The hostname or IP to scan
-    pub hostname: String,
+    /// A space delimited list of hostnames and IPs to trace
+    #[clap(required = true)]
+    pub targets: Vec<String>,
 
     /// Tracing protocol.
     #[clap(arg_enum, short = 'p', long, default_value = "icmp")]
@@ -150,6 +151,10 @@ pub struct Args {
     #[clap(long, short = 'z')]
     pub dns_lookup_as_info: bool,
 
+    /// The maximum number of samples to record per hop.
+    #[clap(long, short = 's', default_value_t = 256)]
+    pub tui_max_samples: usize,
+
     /// Preserve the screen on exit
     #[clap(long)]
     pub tui_preserve_screen: bool,
@@ -175,6 +180,23 @@ pub struct Args {
     pub report_cycles: usize,
 }
 
+/// We only allow multiple targets to be specified for the Tui and for `Icmp` tracing.
+pub fn validate_multi(mode: Mode, protocol: TraceProtocol, targets: &[String]) {
+    match (mode, protocol) {
+        (Mode::Stream | Mode::Pretty | Mode::Markdown | Mode::Csv | Mode::Json, _)
+            if targets.len() > 1 =>
+        {
+            eprintln!("only a single target may be specified for this mode");
+            exit(-1);
+        }
+        (_, TraceProtocol::Tcp | TraceProtocol::Udp) if targets.len() > 1 => {
+            eprintln!("only a single target may be specified for TCP and UDP tracing");
+            exit(-1);
+        }
+        _ => {}
+    }
+}
+
 /// Validate `report_cycles`
 pub fn validate_report_cycles(report_cycles: usize) {
     if report_cycles == 0 {
