@@ -1,10 +1,11 @@
 use crate::backend::Hop;
 use crate::config::{AddressMode, DnsResolveMethod};
+use crate::dns::{DnsEntry, Resolved};
 use crate::{DnsResolver, Trace};
 use chrono::SecondsFormat;
 use crossterm::event::KeyModifiers;
 use crossterm::{
-    event::{self, DisableMouseCapture, Event, KeyCode},
+    event::{self, Event, KeyCode},
     execute,
     terminal::{disable_raw_mode, enable_raw_mode, EnterAlternateScreen, LeaveAlternateScreen},
 };
@@ -15,10 +16,9 @@ use std::io;
 use std::net::IpAddr;
 use std::sync::Arc;
 use std::time::{Duration, SystemTime};
-use trippy::tracing::TracerConfig;
 use tui::layout::{Alignment, Direction, Rect};
 use tui::text::{Span, Spans};
-use tui::widgets::{BarChart, BorderType, Clear, Paragraph, Sparkline, TableState};
+use tui::widgets::{BarChart, BorderType, Clear, Paragraph, Sparkline, TableState, Tabs};
 use tui::{
     backend::{Backend, CrosstermBackend},
     layout::{Constraint, Layout},
@@ -45,66 +45,155 @@ const TABLE_WIDTH: [Constraint; 11] = [
     Constraint::Percentage(5),
 ];
 
+const LAYOUT_WITHOUT_TABS: [Constraint; 3] = [
+    Constraint::Length(5),
+    Constraint::Min(10),
+    Constraint::Length(6),
+];
+
+const LAYOUT_WITH_TABS: [Constraint; 4] = [
+    Constraint::Length(5),
+    Constraint::Length(3),
+    Constraint::Min(10),
+    Constraint::Length(6),
+];
+
+const HELP_LINES: [&str; 14] = [
+    "[up] & [down]    - select hop",
+    "[left] & [right] - select trace",
+    "[esc]            - clear selection",
+    "f                - toggle freeze display",
+    "Ctrl+r           - reset statistics",
+    "Ctrl+k           - flush DNS cache",
+    "i                - show IP only",
+    "n                - show hostname only",
+    "b                - show both IP and hostname",
+    "[ & ]            - expand & collapse hosts",
+    "{ & }            - expand & collapse hosts to max and min",
+    "z                - toggle AS information (if available)",
+    "h                - toggle help",
+    "q                - quit",
+];
+
+/// Information about a `Trace` needed for the Tui.
+#[derive(Debug, Clone)]
+pub struct TuiTraceInfo {
+    pub data: Arc<RwLock<Trace>>,
+    pub target_hostname: String,
+    pub target_addr: IpAddr,
+    pub protocol: String,
+    pub first_ttl: u8,
+    pub max_ttl: u8,
+    pub grace_duration: Duration,
+    pub min_round_duration: Duration,
+}
+
+impl TuiTraceInfo {
+    #[allow(clippy::too_many_arguments)]
+    pub fn new(
+        data: Arc<RwLock<Trace>>,
+        target_hostname: String,
+        target_addr: IpAddr,
+        protocol: String,
+        first_ttl: u8,
+        max_ttl: u8,
+        grace_duration: Duration,
+        min_round_duration: Duration,
+    ) -> Self {
+        Self {
+            data,
+            target_hostname,
+            target_addr,
+            protocol,
+            first_ttl,
+            max_ttl,
+            grace_duration,
+            min_round_duration,
+        }
+    }
+}
+
 /// Tui configuration.
+#[derive(Debug)]
 pub struct TuiConfig {
-    /// The IP address of the target.
-    target_addr: IpAddr,
-    /// The hostname of the target.
-    hostname: String,
     /// Refresh rate.
     refresh_rate: Duration,
     /// Preserve screen on exit.
     preserve_screen: bool,
     /// How to render addresses.
     address_mode: AddressMode,
+    /// Lookup `AS` information.
+    lookup_as_info: bool,
     /// The maximum number of addresses to show per hop.
     max_addrs: Option<u8>,
+    /// The maximum number of samples to record per hop.
+    max_samples: usize,
 }
 
 impl TuiConfig {
     pub fn new(
-        target_addr: IpAddr,
-        hostname: String,
         refresh_rate: Duration,
         preserve_screen: bool,
         address_mode: AddressMode,
+        lookup_as_info: bool,
         max_addrs: Option<u8>,
+        max_samples: usize,
     ) -> Self {
         Self {
-            target_addr,
-            hostname,
             refresh_rate,
             preserve_screen,
             address_mode,
+            lookup_as_info,
             max_addrs,
+            max_samples,
         }
     }
 }
 
 struct TuiApp {
-    tracer_config: TracerConfig,
+    selected_tracer_data: Trace,
+    trace_info: Vec<TuiTraceInfo>,
     tui_config: TuiConfig,
     table_state: TableState,
-    trace: Trace,
+    trace_selected: usize,
     resolver: DnsResolver,
     show_help: bool,
     frozen_start: Option<SystemTime>,
 }
 
 impl TuiApp {
-    fn new(tracer_config: TracerConfig, tui_config: TuiConfig, resolver: DnsResolver) -> Self {
+    fn new(tui_config: TuiConfig, resolver: DnsResolver, trace_info: Vec<TuiTraceInfo>) -> Self {
         Self {
-            table_state: TableState::default(),
-            trace: Trace::default(),
-            resolver,
-            tracer_config,
+            selected_tracer_data: Trace::new(tui_config.max_samples),
+            trace_info,
             tui_config,
+            table_state: TableState::default(),
+            trace_selected: 0,
+            resolver,
             show_help: false,
             frozen_start: None,
         }
     }
-    pub fn next(&mut self) {
-        let hop_count = self.trace.hops().len();
+
+    fn tracer_data(&self) -> &Trace {
+        &self.selected_tracer_data
+    }
+
+    fn snapshot_trace_data(&mut self) {
+        self.selected_tracer_data = self.trace_info[self.trace_selected].data.read().clone();
+    }
+
+    fn clear_trace_data(&mut self) {
+        *self.trace_info[self.trace_selected].data.write() =
+            Trace::new(self.tui_config.max_samples);
+    }
+
+    fn tracer_config(&self) -> &TuiTraceInfo {
+        &self.trace_info[self.trace_selected]
+    }
+
+    fn next_hop(&mut self) {
+        let hop_count = self.tracer_data().hops().len();
         if hop_count == 0 {
             return;
         }
@@ -122,8 +211,8 @@ impl TuiApp {
         self.table_state.select(Some(i));
     }
 
-    pub fn previous(&mut self) {
-        let hop_count = self.trace.hops().len();
+    fn previous_hop(&mut self) {
+        let hop_count = self.tracer_data().hops().len();
         if hop_count == 0 {
             return;
         }
@@ -140,26 +229,75 @@ impl TuiApp {
         self.table_state.select(Some(i));
     }
 
-    pub fn clear(&mut self) {
+    fn next_trace(&mut self) {
+        if self.trace_selected < self.trace_info.len() - 1 {
+            self.trace_selected += 1;
+        }
+    }
+
+    fn previous_trace(&mut self) {
+        if self.trace_selected > 0 {
+            self.trace_selected -= 1;
+        };
+    }
+
+    fn clear(&mut self) {
         self.table_state.select(None);
     }
 
-    pub fn toggle_help(&mut self) {
+    fn toggle_help(&mut self) {
         self.show_help = !self.show_help;
     }
 
-    pub fn toggle_freeze(&mut self) {
+    fn toggle_freeze(&mut self) {
         self.frozen_start = match self.frozen_start {
             None => Some(SystemTime::now()),
             Some(_) => None,
         };
     }
+
+    fn toggle_asinfo(&mut self) {
+        self.tui_config.lookup_as_info = !self.tui_config.lookup_as_info;
+    }
+
+    fn expand_hosts(&mut self) {
+        self.tui_config.max_addrs = match self.tui_config.max_addrs {
+            None => Some(1),
+            Some(i) if i < self.max_hosts() => Some(i + 1),
+            Some(i) => Some(i),
+        }
+    }
+
+    fn contract_hosts(&mut self) {
+        self.tui_config.max_addrs = match self.tui_config.max_addrs {
+            Some(i) if i > 1 => Some(i - 1),
+            _ => None,
+        }
+    }
+
+    fn expand_hosts_max(&mut self) {
+        self.tui_config.max_addrs = Some(self.max_hosts());
+    }
+
+    fn contract_hosts_min(&mut self) {
+        self.tui_config.max_addrs = Some(1);
+    }
+
+    /// The maximum number of hosts per hop for the currently selected trace.
+    fn max_hosts(&self) -> u8 {
+        self.selected_tracer_data
+            .hops()
+            .iter()
+            .map(|h| h.addrs().count())
+            .max()
+            .and_then(|i| u8::try_from(i).ok())
+            .unwrap_or_default()
+    }
 }
 
 /// Run the frontend TUI.
 pub fn run_frontend(
-    trace: &Arc<RwLock<Trace>>,
-    tracer_config: TracerConfig,
+    traces: Vec<TuiTraceInfo>,
     tui_config: TuiConfig,
     resolver: DnsResolver,
 ) -> anyhow::Result<()> {
@@ -169,12 +307,11 @@ pub fn run_frontend(
     let backend = CrosstermBackend::new(stdout);
     let mut terminal = Terminal::new(backend)?;
     let preserve_screen = tui_config.preserve_screen;
-    let res = run_app(&mut terminal, trace, tui_config, tracer_config, resolver);
+    let res = run_app(&mut terminal, traces, tui_config, resolver);
     disable_raw_mode()?;
     if !preserve_screen {
         execute!(terminal.backend_mut(), LeaveAlternateScreen)?;
     }
-    execute!(terminal.backend_mut(), DisableMouseCapture)?;
     terminal.show_cursor()?;
     if let Err(err) = res {
         println!("{:?}", err);
@@ -184,17 +321,16 @@ pub fn run_frontend(
 
 fn run_app<B: Backend>(
     terminal: &mut Terminal<B>,
-    trace: &Arc<RwLock<Trace>>,
+    trace_info: Vec<TuiTraceInfo>,
     tui_config: TuiConfig,
-    tracer_config: TracerConfig,
     resolver: DnsResolver,
 ) -> io::Result<()> {
-    let mut app = TuiApp::new(tracer_config, tui_config, resolver);
+    let mut app = TuiApp::new(tui_config, resolver, trace_info);
     loop {
         if app.frozen_start == None {
-            app.trace = trace.read().clone();
+            app.snapshot_trace_data();
         };
-        terminal.draw(|f| render_all(f, &mut app))?;
+        terminal.draw(|f| render_app(f, &mut app))?;
         if event::poll(app.tui_config.refresh_rate)? {
             if let Event::Key(key) = event::read()? {
                 match (key.code, key.modifiers) {
@@ -204,12 +340,47 @@ fn run_app<B: Backend>(
                     (KeyCode::Char('f'), _) if !app.show_help => app.toggle_freeze(),
                     (KeyCode::Char('r'), KeyModifiers::CONTROL) if !app.show_help => {
                         app.clear();
-                        *trace.write() = Trace::default();
+                        app.clear_trace_data();
                     }
-                    (KeyCode::Down, _) if !app.show_help => app.next(),
-                    (KeyCode::Up, _) if !app.show_help => app.previous(),
+                    (KeyCode::Char('k'), KeyModifiers::CONTROL) if !app.show_help => {
+                        app.resolver.flush();
+                    }
+                    (KeyCode::Down, _) if !app.show_help => app.next_hop(),
+                    (KeyCode::Up, _) if !app.show_help => app.previous_hop(),
                     (KeyCode::Esc, _) if !app.show_help => app.clear(),
                     (KeyCode::Esc, _) if app.show_help => app.toggle_help(),
+                    (KeyCode::Left, _) if !app.show_help => {
+                        app.previous_trace();
+                        app.clear();
+                    }
+                    (KeyCode::Right, _) if !app.show_help => {
+                        app.next_trace();
+                        app.clear();
+                    }
+                    (KeyCode::Char('i'), _) if !app.show_help => {
+                        app.tui_config.address_mode = AddressMode::IP;
+                    }
+                    (KeyCode::Char('n'), _) if !app.show_help => {
+                        app.tui_config.address_mode = AddressMode::Host;
+                    }
+                    (KeyCode::Char('b'), _) if !app.show_help => {
+                        app.tui_config.address_mode = AddressMode::Both;
+                    }
+                    (KeyCode::Char('z'), _) if !app.show_help => {
+                        match app.resolver.config().resolve_method {
+                            DnsResolveMethod::Resolv
+                            | DnsResolveMethod::Google
+                            | DnsResolveMethod::Cloudflare => {
+                                app.toggle_asinfo();
+                                app.resolver.flush();
+                            }
+                            DnsResolveMethod::System => {}
+                        }
+                    }
+                    (KeyCode::Char('{'), _) if !app.show_help => app.contract_hosts_min(),
+                    (KeyCode::Char('}'), _) if !app.show_help => app.expand_hosts_max(),
+                    (KeyCode::Char('['), _) if !app.show_help => app.contract_hosts(),
+                    (KeyCode::Char(']'), _) if !app.show_help => app.expand_hosts(),
                     _ => {}
                 }
             }
@@ -224,6 +395,8 @@ fn run_app<B: Backend>(
 ///  ____________________________________
 /// |               Header               |
 ///  ------------------------------------
+/// |                Tabs                |
+///  ------------------------------------
 /// |                                    |
 /// |                                    |
 /// |                                    |
@@ -237,70 +410,31 @@ fn run_app<B: Backend>(
 ///  ------------------------------------
 ///
 /// Header - the title, configuration, destination, clock and keyboard controls
+/// Tab - a tab for each target being traced (only shown if > 1 target requested)
 /// Hops - a table where each row represents a single hop (time-to-live) in the trace
 /// History - a graph of historic round-trip ping samples for the target host
 /// Frequency - a histogram of sample frequencies by round-trip time for the target host
 ///
 /// On startup a splash screen is shown in place of the hops table, until the completion of the first round.
-fn render_all<B: Backend>(f: &mut Frame<'_, B>, app: &mut TuiApp) {
+fn render_app<B: Backend>(f: &mut Frame<'_, B>, app: &mut TuiApp) {
+    let constraints = if app.trace_info.len() > 1 {
+        LAYOUT_WITH_TABS.as_slice()
+    } else {
+        LAYOUT_WITHOUT_TABS.as_slice()
+    };
     let chunks = Layout::default()
         .direction(Direction::Vertical)
-        .constraints(
-            [
-                Constraint::Percentage(15),
-                Constraint::Percentage(65),
-                Constraint::Percentage(20),
-            ]
-            .as_ref(),
-        )
+        .constraints(constraints.as_ref())
         .split(f.size());
-    let bottom_chunks = Layout::default()
-        .direction(Direction::Horizontal)
-        .constraints([Constraint::Percentage(75), Constraint::Percentage(25)].as_ref())
-        .split(chunks[2]);
     render_header(f, app, chunks[0]);
-    if app.trace.hops().is_empty() {
-        render_splash(f, chunks[1]);
+    if app.trace_info.len() > 1 {
+        render_tabs(f, app, chunks[1]);
+        render_body(f, chunks[2], app);
+        render_footer(f, chunks[3], app);
     } else {
-        render_table(f, app, chunks[1]);
+        render_body(f, chunks[1], app);
+        render_footer(f, chunks[2], app);
     }
-    render_history(f, app, bottom_chunks[0]);
-    render_ping_frequency(f, app, bottom_chunks[1]);
-    if app.show_help {
-        render_help(f);
-    }
-}
-
-/// Render help
-fn render_help<B: Backend>(f: &mut Frame<'_, B>) {
-    let block = Block::default()
-        .title(" Controls ")
-        .title_alignment(Alignment::Center)
-        .borders(Borders::ALL)
-        .style(Style::default().bg(Color::Blue))
-        .border_type(BorderType::Double);
-    let up_down_span = Spans::from(vec![Span::raw("[up]/[down] - select hop")]);
-    let esc_span = Spans::from(vec![Span::raw("[esc]       - clear selection")]);
-    let pause_span = Spans::from(vec![Span::raw("[f]         - toggle freeze display")]);
-    let reset_span = Spans::from(vec![Span::raw("Ctrl+[r]    - reset statistics")]);
-    let help_span = Spans::from(vec![Span::raw("[h]         - toggle help")]);
-    let quit_span = Spans::from(vec![Span::raw("[q]         - quit")]);
-    let control_spans = vec![
-        up_down_span,
-        esc_span,
-        pause_span,
-        reset_span,
-        help_span,
-        quit_span,
-    ];
-    let control = Paragraph::new(control_spans)
-        .style(Style::default())
-        .block(block.clone())
-        .alignment(Alignment::Left);
-    let area = centered_rect(50, 30, f.size());
-    f.render_widget(Clear, area);
-    f.render_widget(block, area);
-    f.render_widget(control, area);
 }
 
 /// Render the title, config, target, clock and keyboard controls.
@@ -324,26 +458,39 @@ fn render_header<B: Backend>(f: &mut Frame<'_, B>, app: &mut TuiApp, rect: Rect)
         .style(Style::default())
         .block(header_block.clone())
         .alignment(Alignment::Right);
+    let protocol = &app.tracer_config().protocol;
+    let dns = format_dns_method(app.resolver.config().resolve_method);
+    let as_info = match app.resolver.config().resolve_method {
+        DnsResolveMethod::System => String::from("n/a"),
+        DnsResolveMethod::Resolv | DnsResolveMethod::Google | DnsResolveMethod::Cloudflare => {
+            if app.tui_config.lookup_as_info {
+                String::from("on")
+            } else {
+                String::from("off")
+            }
+        }
+    };
+    let interval = humantime::format_duration(app.tracer_config().min_round_duration);
+    let grace = humantime::format_duration(app.tracer_config().grace_duration);
+    let first_ttl = app.tracer_config().first_ttl;
+    let max_ttl = app.tracer_config().max_ttl;
+    let max_hosts = app
+        .tui_config
+        .max_addrs
+        .map_or_else(|| String::from("auto"), |m| m.to_string());
     let left_spans = vec![
         Spans::from(vec![
             Span::styled("Target: ", Style::default().add_modifier(Modifier::BOLD)),
             Span::raw(format!(
                 "{} ({})",
-                app.tui_config.hostname, app.tui_config.target_addr
+                app.tracer_config().target_hostname,
+                app.tracer_config().target_addr
             )),
         ]),
         Spans::from(vec![
             Span::styled("Config: ", Style::default().add_modifier(Modifier::BOLD)),
-            Span::raw(format!(
-                "protocol={} dns={} interval={} grace={} start-ttl={} max-ttl={}",
-                app.tracer_config.protocol,
-                format_dns_method(app.resolver.config().resolve_method),
-                humantime::format_duration(app.tracer_config.min_round_duration),
-                humantime::format_duration(app.tracer_config.grace_duration),
-                app.tracer_config.first_ttl.0,
-                app.tracer_config.max_ttl.0
-            )),
-        ]),
+            Span::raw(format!("protocol={} dns={} as-info={} interval={} grace={} start-ttl={} max-ttl={} max-hosts={}",
+                              protocol, dns, as_info, interval, grace, first_ttl, max_ttl, max_hosts))]),
         Spans::from(vec![
             Span::styled("Status: ", Style::default().add_modifier(Modifier::BOLD)),
             Span::raw(if let Some(start) = app.frozen_start {
@@ -356,7 +503,10 @@ fn render_header<B: Backend>(f: &mut Frame<'_, B>, app: &mut TuiApp, rect: Rect)
             } else {
                 String::from("Running")
             }),
-            Span::raw(format!(", discovered {} hops", app.trace.hops().len())),
+            Span::raw(format!(
+                ", discovered {} hops",
+                app.tracer_data().hops().len()
+            )),
         ]),
     ];
 
@@ -378,6 +528,47 @@ fn format_dns_method(resolve_method: DnsResolveMethod) -> String {
     }
 }
 
+/// Render the tabs, one per trace.
+fn render_tabs<B: Backend>(f: &mut Frame<'_, B>, app: &mut TuiApp, rect: Rect) {
+    let tabs_block = Block::default()
+        .title("Traces")
+        .title_alignment(Alignment::Left)
+        .borders(Borders::ALL)
+        .border_type(BorderType::Rounded)
+        .style(Style::default());
+    let titles: Vec<_> = app
+        .trace_info
+        .iter()
+        .map(|trace| {
+            Spans::from(Span::styled(
+                &trace.target_hostname,
+                Style::default().fg(Color::Green),
+            ))
+        })
+        .collect();
+    let tabs = Tabs::new(titles)
+        .block(tabs_block.clone())
+        .select(app.trace_selected)
+        .style(Style::default())
+        .highlight_style(
+            Style::default()
+                .add_modifier(Modifier::BOLD)
+                .bg(Color::Black),
+        );
+    f.render_widget(tabs, rect);
+}
+
+/// Render the body.
+///
+/// This is the table of hop data or, if there is no data, the spash screen.
+fn render_body<B: Backend>(f: &mut Frame<'_, B>, rec: Rect, app: &mut TuiApp) {
+    if app.tracer_data().hops().is_empty() {
+        render_splash(f, rec);
+    } else {
+        render_table(f, app, rec);
+    }
+}
+
 /// Render the splash screen.
 ///
 /// This is shown on startup whilst we await the first round of data to be available.
@@ -426,12 +617,13 @@ fn render_splash<B: Backend>(f: &mut Frame<'_, B>, rect: Rect) {
 fn render_table<B: Backend>(f: &mut Frame<'_, B>, app: &mut TuiApp, rect: Rect) {
     let header = render_table_header();
     let selected_style = Style::default().add_modifier(Modifier::REVERSED);
-    let rows = app.trace.hops().iter().map(|hop| {
+    let rows = app.tracer_data().hops().iter().map(|hop| {
         render_table_row(
             hop,
-            &mut app.resolver,
-            app.trace.is_target(hop),
+            &app.resolver,
+            app.tracer_data().is_target(hop),
             app.tui_config.address_mode,
+            app.tui_config.lookup_as_info,
             app.tui_config.max_addrs,
         )
     });
@@ -462,13 +654,14 @@ fn render_table_header() -> Row<'static> {
 /// Render a single row in the table of hops.
 fn render_table_row(
     hop: &Hop,
-    dns: &mut DnsResolver,
+    dns: &DnsResolver,
     is_target: bool,
     address_mode: AddressMode,
+    lookup_as_info: bool,
     max_addr: Option<u8>,
 ) -> Row<'static> {
     let ttl_cell = render_ttl_cell(hop);
-    let hostname_cell = render_hostname_cell(hop, dns, address_mode, max_addr);
+    let hostname_cell = render_hostname_cell(hop, dns, address_mode, lookup_as_info, max_addr);
     let loss_pct_cell = render_loss_pct_cell(hop);
     let total_sent_cell = render_total_sent_cell(hop);
     let total_recv_cell = render_total_recv_cell(hop);
@@ -524,22 +717,59 @@ fn render_avg_cell(hop: &Hop) -> Cell<'static> {
 
 fn render_hostname_cell(
     hop: &Hop,
-    dns: &mut DnsResolver,
+    dns: &DnsResolver,
     address_mode: AddressMode,
+    lookup_as_info: bool,
     max_addr: Option<u8>,
 ) -> Cell<'static> {
+    /// Format a `DnsEntry` with or without `AS` information (if available)
+    fn format_dns_entry(dns_entry: DnsEntry, lookup_as_info: bool) -> String {
+        match dns_entry {
+            DnsEntry::Resolved(Resolved::Normal(_, hosts)) => hosts.join(" "),
+            DnsEntry::Resolved(Resolved::WithAsInfo(_, hosts, asinfo)) => {
+                if lookup_as_info {
+                    format!("AS{} {}", asinfo.asn, hosts.join(" "))
+                } else {
+                    hosts.join(" ")
+                }
+            }
+            DnsEntry::Pending(ip) | DnsEntry::NotFound(ip) => format!("{}", ip),
+            DnsEntry::Failed(ip) => format!("Failed: {}", ip),
+            DnsEntry::Timeout(ip) => format!("Timeout: {}", ip),
+        }
+    }
+    /// Perform a reverse DNS lookup for an address and format the result.
     fn format_address(
         addr: &IpAddr,
         freq: usize,
         hop: &Hop,
-        dns: &mut DnsResolver,
+        dns: &DnsResolver,
         address_mode: AddressMode,
+        lookup_as_info: bool,
     ) -> String {
         let addr_fmt = match address_mode {
             AddressMode::IP => addr.to_string(),
-            AddressMode::Host => dns.reverse_lookup(*addr).to_string(),
-            AddressMode::Both => format!("{} ({})", dns.reverse_lookup(*addr), addr),
+            AddressMode::Host => {
+                if lookup_as_info {
+                    let entry = dns.reverse_lookup_with_asinfo(*addr);
+                    format_dns_entry(entry, true)
+                } else {
+                    let entry = dns.reverse_lookup(*addr);
+                    format_dns_entry(entry, false)
+                }
+            }
+            AddressMode::Both => {
+                let hostname = if lookup_as_info {
+                    let entry = dns.reverse_lookup_with_asinfo(*addr);
+                    format_dns_entry(entry, true)
+                } else {
+                    let entry = dns.reverse_lookup(*addr);
+                    format_dns_entry(entry, false)
+                };
+                format!("{} ({})", hostname, addr)
+            }
         };
+
         if hop.addr_count() > 1 {
             format!(
                 "{} [{:.1}%]",
@@ -555,14 +785,18 @@ fn render_hostname_cell(
         match max_addr {
             None => hop
                 .addrs_with_counts()
-                .map(|(addr, &freq)| format_address(addr, freq, hop, dns, address_mode))
+                .map(|(addr, &freq)| {
+                    format_address(addr, freq, hop, dns, address_mode, lookup_as_info)
+                })
                 .join("\n"),
             Some(max_addr) => hop
                 .addrs_with_counts()
                 .sorted_unstable_by_key(|(_, &cnt)| cnt)
                 .rev()
                 .take(max_addr as usize)
-                .map(|(addr, &freq)| format_address(addr, freq, hop, dns, address_mode))
+                .map(|(addr, &freq)| {
+                    format_address(addr, freq, hop, dns, address_mode, lookup_as_info)
+                })
                 .join("\n"),
         }
     } else {
@@ -613,12 +847,27 @@ fn render_status_cell(hop: &Hop, is_target: bool) -> Cell<'static> {
     })
 }
 
+/// Render the footer.
+///
+/// This contains the history and frequency charts.
+fn render_footer<B: Backend>(f: &mut Frame<'_, B>, rec: Rect, app: &mut TuiApp) {
+    let bottom_chunks = Layout::default()
+        .direction(Direction::Horizontal)
+        .constraints([Constraint::Percentage(75), Constraint::Percentage(25)].as_ref())
+        .split(rec);
+    render_history(f, app, bottom_chunks[0]);
+    render_ping_frequency(f, app, bottom_chunks[1]);
+    if app.show_help {
+        render_help(f);
+    }
+}
+
 /// Render the ping history for the final hop which is typically the target.
 fn render_history<B: Backend>(f: &mut Frame<'_, B>, app: &mut TuiApp, rect: Rect) {
-    let target_hop = app
-        .table_state
-        .selected()
-        .map_or_else(|| app.trace.target_hop(), |s| &app.trace.hops()[s]);
+    let target_hop = app.table_state.selected().map_or_else(
+        || app.tracer_data().target_hop(),
+        |s| &app.tracer_data().hops()[s],
+    );
     let data = target_hop
         .samples()
         .iter()
@@ -639,10 +888,10 @@ fn render_history<B: Backend>(f: &mut Frame<'_, B>, app: &mut TuiApp, rect: Rect
 
 /// Render a histogram of ping frequencies.
 fn render_ping_frequency<B: Backend>(f: &mut Frame<'_, B>, app: &mut TuiApp, rect: Rect) {
-    let target_hop = app
-        .table_state
-        .selected()
-        .map_or_else(|| app.trace.target_hop(), |s| &app.trace.hops()[s]);
+    let target_hop = app.table_state.selected().map_or_else(
+        || app.tracer_data().target_hop(),
+        |s| &app.tracer_data().hops()[s],
+    );
     let freq_data = sample_frequency(target_hop.samples());
     let freq_data_ref: Vec<_> = freq_data.iter().map(|(b, c)| (b.as_str(), *c)).collect();
     let barchart = BarChart::default()
@@ -664,6 +913,25 @@ fn render_ping_frequency<B: Backend>(f: &mut Frame<'_, B>, app: &mut TuiApp, rec
     f.render_widget(barchart, rect);
 }
 
+/// Render help
+fn render_help<B: Backend>(f: &mut Frame<'_, B>) {
+    let block = Block::default()
+        .title(" Controls ")
+        .title_alignment(Alignment::Center)
+        .borders(Borders::ALL)
+        .style(Style::default().bg(Color::Blue))
+        .border_type(BorderType::Double);
+    let control_spans: Vec<_> = HELP_LINES.iter().map(|&line| Spans::from(line)).collect();
+    let control = Paragraph::new(control_spans)
+        .style(Style::default())
+        .block(block.clone())
+        .alignment(Alignment::Left);
+    let area = centered_rect(50, 40, f.size());
+    f.render_widget(Clear, area);
+    f.render_widget(block, area);
+    f.render_widget(control, area);
+}
+
 fn centered_rect(percent_x: u16, percent_y: u16, r: Rect) -> Rect {
     let popup_layout = Layout::default()
         .direction(Direction::Vertical)
