@@ -9,4 +9,4 @@ mod util;
 pub use config::{TracerConfig, TracerProtocol};
 pub use net::TracerChannel;
 pub use probe::{IcmpPacketType, Probe, ProbeStatus};
-pub use tracer::Tracer;
+pub use tracer::{Tracer, TracerRound};
