@@ -9,6 +9,8 @@ pub type TraceResult<T> = Result<T, TracerError>;
 pub enum TracerError {
     #[error("invalid packet size: {0}")]
     InvalidPacketSize(usize),
+    #[error("unknown default interface")]
+    UnknownDefaultInterface,
     #[error("invalid config: {0}")]
     BadConfig(String),
     #[error("missing required field: {0}")]
