@@ -2,11 +2,42 @@ use self::state::TracerState;
 use crate::tracing::error::TraceResult;
 use crate::tracing::net::{Network, ProbeResponse};
 use crate::tracing::types::{MaxInflight, MaxRounds, Sequence, TimeToLive, TraceId};
-use crate::tracing::TracerConfig;
+use crate::tracing::TracerProtocol;
 use crate::tracing::{IcmpPacketType, ProbeStatus};
-use crate::tracing::{Probe, TracerProtocol};
+use crate::tracing::{Probe, TracerConfig};
 use std::time::{Duration, SystemTime};
 
+/// The output from a round of tracing.
+#[derive(Debug, Clone)]
+pub struct TracerRound<'a> {
+    /// The state of all `Probe` that were sent in the round.
+    pub probes: &'a [Probe],
+    /// The largest time-to-live (ttl) for which we received a reply in the round.
+    pub largest_ttl: TimeToLive,
+    /// Indicates what triggered the completion of the tracing round.
+    pub reason: CompletionReason,
+}
+
+impl<'a> TracerRound<'a> {
+    #[must_use]
+    pub fn new(probes: &'a [Probe], largest_ttl: TimeToLive, reason: CompletionReason) -> Self {
+        Self {
+            probes,
+            largest_ttl,
+            reason,
+        }
+    }
+}
+
+/// Indicates what triggered the completion of the tracing round.
+#[derive(Debug, Copy, Clone)]
+pub enum CompletionReason {
+    /// The round ended because the target was found.
+    TargetFound,
+    /// The round ended because the time exceeded the configured maximum round time.
+    RoundTimeLimitExceeded,
+}
+
 /// Trace a path to a target.
 #[derive(Debug, Clone)]
 pub struct Tracer<F> {
@@ -24,7 +55,7 @@ pub struct Tracer<F> {
     publish: F,
 }
 
-impl<F: Fn(&Probe)> Tracer<F> {
+impl<F: Fn(&TracerRound<'_>)> Tracer<F> {
     pub fn new(config: &TracerConfig, publish: F) -> Self {
         Self {
             protocol: config.protocol,
@@ -85,9 +116,10 @@ impl<F: Fn(&Probe)> Tracer<F> {
     /// Read and process the next incoming `ICMP` packet.
     ///
     /// We allow multiple probes to be in-flight at any time and we cannot guaranteed that responses will be
-    /// received in-order.  We therefore maintain a circular buffer which holds details of each `Probe` which is
-    /// indexed by a sequence number (modulo the buffer size).  The sequence number is set in the outgoing `ICMP`
-    /// `EchoRequest` (or `UDP` / `TCP`) packet and returned in both the `TimeExceeded` and `EchoReply` responses.
+    /// received in-order.  We therefore maintain a buffer which holds details of each `Probe` which is
+    /// indexed by the offset of the sequence number from the sequence number at the beginning of the round.  The
+    /// sequence number is set in the outgoing `ICMP` `EchoRequest` (or `UDP` / `TCP`) packet and returned in both
+    /// the `TimeExceeded` and `EchoReply` responses.
     ///
     /// Each incoming `ICMP` packet contains the original `ICMP` `EchoRequest` packet from which we can read the
     /// `identifier` that we set which we can now validate to ensure we only process responses which correspond to
@@ -109,9 +141,7 @@ impl<F: Fn(&Probe)> Tracer<F> {
                 let received = data.recv;
                 let ip = data.addr;
                 let trace_id = TraceId::from(data.identifier);
-                if (self.trace_identifier == trace_id || trace_id == TraceId::from(0))
-                    && st.in_round(sequence)
-                {
+                if self.check_trace_id(trace_id) && st.in_round(sequence) {
                     let probe = st
                         .probe_at(sequence)
                         .with_status(ProbeStatus::Complete)
@@ -126,14 +156,14 @@ impl<F: Fn(&Probe)> Tracer<F> {
                 let received = data.recv;
                 let ip = data.addr;
                 let trace_id = TraceId::from(data.identifier);
-                if self.trace_identifier == trace_id && st.in_round(sequence) {
+                if self.check_trace_id(trace_id) && st.in_round(sequence) {
                     let probe = st
                         .probe_at(sequence)
                         .with_status(ProbeStatus::Complete)
                         .with_icmp_packet_type(IcmpPacketType::Unreachable)
                         .with_host(ip)
                         .with_received(received);
-                    st.update_probe(sequence, probe, received, false);
+                    st.update_probe(sequence, probe, received, true);
                 }
             }
             Some(ProbeResponse::EchoReply(data)) => {
@@ -141,7 +171,7 @@ impl<F: Fn(&Probe)> Tracer<F> {
                 let received = data.recv;
                 let ip = data.addr;
                 let trace_id = TraceId::from(data.identifier);
-                if self.trace_identifier == trace_id && st.in_round(sequence) {
+                if self.check_trace_id(trace_id) && st.in_round(sequence) {
                     let probe = st
                         .probe_at(sequence)
                         .with_status(ProbeStatus::Complete)
@@ -162,17 +192,17 @@ impl<F: Fn(&Probe)> Tracer<F> {
     ///
     /// 1 - the round has exceed the minimum round duration AND
     /// 2 - the duration since the last packet was received exceeds the grace period AND
-    /// 2 - either:
+    /// 3 - either:
     ///     A - the target has been found OR
     ///     B - the target has not been found and the round has exceeded the maximum round duration
     fn update_round(&self, st: &mut TracerState) {
         let now = SystemTime::now();
         let round_duration = now.duration_since(st.round_start()).unwrap_or_default();
-        if round_duration > self.min_round_duration
-            && exceeds(st.received_time(), now, self.grace_duration)
-            && st.target_found()
-            || round_duration > self.max_round_duration
-        {
+        let round_min = round_duration > self.min_round_duration;
+        let grace_exceeded = exceeds(st.received_time(), now, self.grace_duration);
+        let round_max = round_duration > self.max_round_duration;
+        let target_found = st.target_found();
+        if round_min && grace_exceeded && target_found || round_max {
             self.publish_trace(st);
             st.advance_round(self.first_ttl);
         }
@@ -183,28 +213,31 @@ impl<F: Fn(&Probe)> Tracer<F> {
     /// If the round completed without receiving an `EchoReply` from the target host then we also publish the next
     /// `Probe` which is assumed to represent the TTL of the target host.
     fn publish_trace(&self, state: &TracerState) {
-        let round_size = if let Some(target_ttl) = state.target_ttl() {
-            // If we started at ttl N and found the target at ttl M then the round contains M - N + 1 entries
-            target_ttl.0 - self.first_ttl.0 + 1
+        let max_received_ttl = if let Some(target_ttl) = state.target_ttl() {
+            target_ttl
         } else {
-            // If we did not receive any responses then the round is size 0
-            // If we started at ttl N and received a max ttl response M then the round contains M - N + 2 entries,
-            // where the 'extra' entry represents the next ttl which did not receive a response.  This is capped by the
-            // maximum allowed round size and so the largest ttl may not be the 'extra' one.
-            state.max_received_ttl().map_or(0, |max_received_ttl| {
-                let size = max_received_ttl.0.saturating_sub(self.first_ttl.0) + 1;
-                let max_allowed = self.max_ttl.0 - self.first_ttl.0;
-                size.min(max_allowed) + 1
-            })
+            state
+                .max_received_ttl()
+                .map_or(TimeToLive(0), |max_received_ttl| {
+                    let max_sent_ttl = state.ttl() - TimeToLive(1);
+                    max_sent_ttl.min(max_received_ttl + TimeToLive(1))
+                })
         };
-        state
-            .probes()
-            .take(usize::from(round_size))
-            .for_each(|probe| {
-                debug_assert_eq!(probe.round, state.round());
-                debug_assert_ne!(probe.ttl.0, 0);
-                (self.publish)(probe);
-            });
+        let probes = state.probes();
+        let largest_ttl = max_received_ttl;
+        let reason = if state.target_found() {
+            CompletionReason::TargetFound
+        } else {
+            CompletionReason::RoundTimeLimitExceeded
+        };
+        (self.publish)(&TracerRound::new(probes, largest_ttl, reason));
+    }
+
+    /// Check if the `TraceId` matches the expected value for this tracer.
+    ///
+    /// A special value of `0` is accepted for `udp` and `tcp` which do not have an identifier.
+    fn check_trace_id(&self, trace_id: TraceId) -> bool {
+        self.trace_identifier == trace_id || trace_id == TraceId::from(0)
     }
 }
 
@@ -217,7 +250,7 @@ mod state {
     use crate::tracing::Probe;
     use std::time::SystemTime;
 
-    /// The maximum number of `Probe` entries in the circular buffer.
+    /// The maximum number of `Probe` entries in the buffer.
     ///
     /// This is effectively also the maximum number of time-to-live (TTL) we can support.  We only ever send TTL in
     /// the range 1..255 so this is technically one larger than we need.
@@ -276,27 +309,21 @@ mod state {
             }
         }
 
-        /// Get an iterator over the `Probe` in the current round.
-        pub fn probes(&self) -> impl Iterator<Item = &Probe> + '_ {
-            self.buffer
-                .iter()
-                .cycle()
-                .skip(usize::from(self.round_sequence % BUFFER_SIZE))
+        /// Get a slice of `Probe` for the current round.
+        pub fn probes(&self) -> &[Probe] {
+            let round_size = self.sequence - self.round_sequence;
+            &self.buffer[..round_size.0 as usize]
         }
 
         /// Get the `Probe` for `sequence`
         pub fn probe_at(&self, sequence: Sequence) -> Probe {
-            self.buffer[usize::from(sequence % BUFFER_SIZE)]
+            self.buffer[usize::from(sequence - self.round_sequence)]
         }
 
         pub const fn ttl(&self) -> TimeToLive {
             self.ttl
         }
 
-        pub const fn round(&self) -> Round {
-            self.round
-        }
-
         pub const fn round_start(&self) -> SystemTime {
             self.round_start
         }
@@ -336,7 +363,7 @@ mod state {
         /// us to use a `u8`.
         pub fn next_probe(&mut self) -> Probe {
             let probe = Probe::new(self.sequence, self.ttl, self.round, SystemTime::now());
-            self.buffer[usize::from(self.sequence % BUFFER_SIZE)] = probe;
+            self.buffer[usize::from(self.sequence - self.round_sequence)] = probe;
             debug_assert!(self.ttl < TimeToLive(u8::MAX));
             self.ttl += TimeToLive::from(1);
             debug_assert!(self.sequence < Sequence(u16::MAX));
@@ -373,7 +400,7 @@ mod state {
                 }
                 _ => {}
             }
-            self.buffer[usize::from(sequence % BUFFER_SIZE)] = probe;
+            self.buffer[usize::from(sequence - self.round_sequence)] = probe;
             self.max_received_ttl = match self.max_received_ttl {
                 Some(max_received_ttl) => Some(max_received_ttl.max(probe.ttl)),
                 None => Some(probe.ttl),
@@ -490,18 +517,10 @@ mod state {
 
             // Validate the probes() iterator returns returns only a single probe
             {
-                let mut probe_iter = state.probes();
+                let mut probe_iter = state.probes().iter();
                 let probe_next1 = *probe_iter.next().unwrap();
                 assert_eq!(probe_1_fetch, probe_next1);
-                let probe_next2 = *probe_iter.next().unwrap();
-                assert_eq!(probe_next2.sequence, Sequence(0));
-                assert_eq!(probe_next2.ttl, TimeToLive(0));
-                assert_eq!(probe_next2.round, Round(0));
-                assert_eq!(probe_next2.received, None);
-                assert_eq!(probe_next2.host, None);
-                assert_eq!(probe_next2.sent.is_some(), false);
-                assert_eq!(probe_next2.status, ProbeStatus::NotSent);
-                assert_eq!(probe_next2.icmp_packet_type, None);
+                assert_eq!(None, probe_iter.next());
             }
 
             // Advance to the next round
@@ -564,7 +583,7 @@ mod state {
 
             // Validate the probes() iterator returns the two probes in the states we expect
             {
-                let mut probe_iter = state.probes();
+                let mut probe_iter = state.probes().iter();
                 let probe_next1 = *probe_iter.next().unwrap();
                 assert_eq!(probe_2_recv, probe_next1);
                 let probe_next2 = *probe_iter.next().unwrap();
@@ -595,7 +614,7 @@ mod state {
 
             // Validate the probes() iterator returns the two probes in the states we expect
             {
-                let mut probe_iter = state.probes();
+                let mut probe_iter = state.probes().iter();
                 let probe_next1 = *probe_iter.next().unwrap();
                 assert_eq!(probe_2_recv, probe_next1);
                 let probe_next2 = *probe_iter.next().unwrap();
@@ -618,7 +637,7 @@ mod state {
 
             // Validate the probes()
             {
-                let mut iter = state.probes();
+                let mut iter = state.probes().iter();
                 assert_eq!(iter.next().unwrap().sequence, Sequence(65278));
                 iter.take(BUFFER_SIZE as usize - 1)
                     .for_each(|p| assert_eq!(p.sequence, Sequence(0)));
@@ -636,7 +655,7 @@ mod state {
 
             // Validate the probes() again
             {
-                let mut iter = state.probes();
+                let mut iter = state.probes().iter();
                 assert_eq!(iter.next().unwrap().sequence, Sequence(65278));
                 iter.take(BUFFER_SIZE as usize - 1)
                     .for_each(|p| assert_eq!(p.sequence, Sequence(0)));
